@@ -5,7 +5,7 @@ import struct
 import logging
 import binascii
 import datetime
-from cStringIO import StringIO
+from StringIO import StringIO as _StringIO
 from OpenSSL import SSL, crypto
 from twisted.internet import reactor, defer
 from twisted.internet.protocol import (
@@ -27,6 +27,18 @@ FEEDBACK_SERVER_PORT = 2196
 
 app_ids = {} # {'app_id': APNSService()}
 
+class StringIO(_StringIO):
+  """Add context management protocol to StringIO
+      ie: http://bugs.python.org/issue1286
+  """
+  
+  def __enter__(self):
+    if self.closed:
+      raise ValueError('I/O operation on closed file')
+    return self
+  
+  def __exit__(self, exc, value, tb):
+    self.close()
 
 class IAPNSService(Interface):
     """ Interface for APNS """
@@ -85,8 +97,8 @@ class APNSFeedbackHandler(LineReceiver):
 
   def connectionLost(self, reason):
     log.msg('feedbackHandler connectionLost %s' % reason)
-    self.deferred.callback(self.io.getValue())
-    io.close()
+    self.deferred.callback(self.io.getvalue())
+    self.io.close()
 
 
 class APNSFeedbackClientFactory(ClientFactory):
@@ -99,7 +111,7 @@ class APNSFeedbackClientFactory(ClientFactory):
     p = self.protocol()
     p.factory = self
     p.deferred = self.deferred
-    p.io = StringIO
+    p.io = StringIO()
     p.setRawMode()
     return p
   
@@ -108,11 +120,11 @@ class APNSFeedbackClientFactory(ClientFactory):
   
   def clientConnectionLost(self, connector, reason):
     log.msg('APNSFeedbackClientFactory clientConnectionLost reason=%s' % reason)
-    ReconnectingClientFactory.clientConnectionLost(self, connector, reason)
+    ClientFactory.clientConnectionLost(self, connector, reason)
   
   def clientConnectionFailed(self, connector, reason):
     log.msg('APNSFeedbackClientFactory clientConnectionFailed reason=%s' % reason)
-    ReconnectingClientFactory.clientConnectionLost(self, connector, reason)
+    ClientFactory.clientConnectionLost(self, connector, reason)
 
 
 class APNSClientFactory(ReconnectingClientFactory):
@@ -121,7 +133,7 @@ class APNSClientFactory(ReconnectingClientFactory):
   def __init__(self):
     self.clientProtocol = None
     self.deferred = defer.Deferred()
-    self.deferred.addErrback(log_errback('APNSClientFactory__init__'))
+    self.deferred.addErrback(log_errback('APNSClientFactory __init__'))
   
   def addClient(self, p):
     self.clientProtocol = p
@@ -130,7 +142,7 @@ class APNSClientFactory(ReconnectingClientFactory):
   def removeClient(self, p):
     self.clientProtocol = None
     self.deferred = defer.Deferred()
-    self.deferred.addErrback(log_errback('APNSClientFactoryremoveClient'))
+    self.deferred.addErrback(log_errback('APNSClientFactory removeClient'))
   
   def startedConnecting(self, connector):
     log.msg('APNSClientFactory startedConnecting')
@@ -154,16 +166,18 @@ class APNSService(service.Service):
   """ A Service that sends notifications and receives 
   feedback from the Apple Push Notification Service
   """
+  
   implements(IAPNSService)
   clientProtocolFactory = APNSClientFactory
   feedbackProtocolFactory = APNSFeedbackClientFactory
   
-  def __init__(self, cert_path, environment):
+  def __init__(self, cert_path, environment, timeout=15):
     log.msg('APNSService __init__')
     self.factory = None
     self.environment = environment
     self.cert_path = cert_path
     self.raw_mode = False
+    self.timeout = timeout
   
   def getContextFactory(self):
     return APNSClientContextFactory(self.cert_path)
@@ -182,10 +196,19 @@ class APNSService(service.Service):
     client = self.factory.clientProtocol
     if client:
       return client.sendMessage(notifications)
-    else:
+    else:      
       d = self.factory.deferred
+      timeout = reactor.callLater(self.timeout, 
+        lambda: d.called or d.errback(
+          Exception('Notification timed out after %i seconds' % self.timeout)))
+      def cancel_timeout(r):
+        try: timeout.cancel()
+        except: pass
+        return r
+      
       d.addCallback(lambda p: p.sendMessage(notifications))
       d.addErrback(log_errback('apns-service-write'))
+      d.addBoth(cancel_timeout)
       return d
   
   def read(self):
@@ -199,6 +222,16 @@ class APNSService(service.Service):
       context = self.getContextFactory()
       reactor.connectSSL(server, port, factory, context)
       factory.deferred.addErrback(log_errback('apns-feedback-read'))
+      
+      timeout = reactor.callLater(self.timeout,
+        lambda: factory.deferred.called or factory.deferred.errback(
+          Exception('Feedbcak fetch timed out after %i seconds' % self.timeout)))
+      def cancel_timeout(r):
+        try: timeout.cancel()
+        except: pass
+        return r
+      
+      factory.deferred.addBoth(cancel_timeout)
     except Exception, e:
       log.err('APNService feedback error initializing: %s' % str(e))
       raise
@@ -215,7 +248,7 @@ class APNSServer(xmlrpc.XMLRPC):
       raise xmlrpc.Fault(404, 'The app_id specified has not been provisioned.')
     return self.app_ids[app_id]
   
-  def xmlrpc_provision(self, app_id, path_to_cert_or_cert, environment):
+  def xmlrpc_provision(self, app_id, path_to_cert_or_cert, environment, timeout=15):
     """ Starts an APNSService for the this app_id and keeps it running
     
       Arguments:
@@ -223,12 +256,18 @@ class APNSServer(xmlrpc.XMLRPC):
           path_to_cert_or_cert   absolute path to the APNS SSL cert or a 
                                  string containing the .pem file
           environment            either 'sandbox' or 'production'
+          timeout                seconds to timeout connection attempts
+                                 to the APNS server
       Returns:
           None
     """
     
+    if environment not in ('sandbox', 'production'):
+      raise xmlrpc.Fault(401, 'Invalid environment provided `%s`. Valid '
+                              'environments are `sandbox` and `production`' % (
+                              environment,))
     if not app_id in self.app_ids:
-      self.app_ids[app_id] = APNSService(path_to_cert_or_cert, environment)
+      self.app_ids[app_id] = APNSService(path_to_cert_or_cert, environment, timeout)
   
   def xmlrpc_notify(self, app_id, token_or_token_list, aps_dict_or_list):
     """ Sends push notifications to the Apple APNS server. Multiple 
@@ -243,12 +282,20 @@ class APNSServer(xmlrpc.XMLRPC):
           None
     """
     
-    self.apns_service(app_id).write(
+    d = self.apns_service(app_id).write(
       encode_notifications(
         [t.replace(' ', '') for t in token_or_token_list] 
           if (type(token_or_token_list) is list)
           else token_or_token_list.replace(' ', ''),
         aps_dict_or_list))
+    if d:
+      def _finish_err(r):
+        # so far, the only error that could really become of this
+        # request is a timeout, since APNS simply terminates connectons
+        # that are made unsuccessfully, which twisted will try endlessly
+        # to reconnect to, we timeout and notifify the client
+        raise xmlrpc.Fault(500, 'Connection to the APNS server could not be made.')
+      return d.addCallbacks(lambda r: None, _finish_err)
   
   def xmlrpc_feedback(self, app_id):
     """ Queries the Apple APNS feedback server for inactive app tokens. Returns
